"""Welcome to SiPANNs, silicon photonics with artificial neural networks,
documentation.

We leverage various machine learning techniques to simulate integrated
photonic device circuits.
"""

__version__ = "1.3.1"
__author__ = "Easton Potokar, Alec Hammond, R Scott Collings <eastonpost@byu.edu>"
__all__ = []
